@@ -1,24 +1,32 @@
-use cargo::{Config, core::manifest::ManifestMetadata, core::registry::PackageRegistry,
-            core::{Dependency, EitherManifest, FeatureValue, Manifest, Package, PackageId,
-                   Registry, Source, SourceId, Summary, Target, TargetKind},
-            sources::registry::RegistrySource, util::{FileLock, toml::read_manifest}};
+use cargo::{
+    core::manifest::ManifestMetadata,
+    core::registry::PackageRegistry,
+    core::{
+        Dependency, EitherManifest, FeatureValue, Manifest, Package, PackageId, Registry, Source,
+        SourceId, Summary, Target, TargetKind,
+    },
+    sources::registry::RegistrySource,
+    util::{toml::read_manifest, FileLock},
+    Config,
+};
 use failure::Error;
 use filetime::{set_file_times, FileTime};
-use glob::Pattern;
-use semver::Version;
 use flate2::read::GzDecoder;
+use glob::Pattern;
+use regex::Regex;
+use semver::Version;
 use tar::Archive;
 use tempfile;
-use regex::Regex;
 
 use std;
-use std::collections::BTreeMap;
-use std::hash::{Hash, Hasher};
-use std::path::{Path, PathBuf};
-use std::io::{self, Read, Write};
+use std::collections::{BTreeMap, HashSet};
 use std::fs;
+use std::hash::{Hash, Hasher};
+use std::io::{self, Read, Write};
+use std::path::{Path, PathBuf};
 
 use errors::*;
+use util::vec_opt_iter;
 
 pub struct CrateInfo {
     package: Package,
@@ -26,6 +34,8 @@ pub struct CrateInfo {
     crate_file: FileLock,
     config: Config,
     source_id: SourceId,
+    excludes: Vec<Pattern>,
+    includes: Vec<Pattern>,
 }
 
 fn hash<H: Hash>(hashable: &H) -> u64 {
@@ -48,14 +58,16 @@ fn traverse_depth<'a>(map: &BTreeMap<&'a str, Vec<&'a str>>, key: &'a str) -> Ve
 
 fn fetch_candidates(registry: &mut PackageRegistry, dep: &Dependency) -> Result<Vec<Summary>> {
     let mut summaries = registry.query_vec(dep, false)?;
-    summaries.sort_by(|a, b| b.package_id().partial_cmp(a.package_id()).unwrap());
+    summaries.sort_by(|a, b| b.package_id().partial_cmp(&a.package_id()).unwrap());
     Ok(summaries)
 }
 
 pub fn update_crates_io() -> Result<()> {
     let config = Config::default()?;
+    let _lock = config.acquire_package_cache_lock()?;
     let source_id = SourceId::crates_io(&config)?;
-    let mut r = RegistrySource::remote(&source_id, &config);
+    let yanked_whitelist = HashSet::new();
+    let mut r = RegistrySource::remote(source_id, &yanked_whitelist, &config);
     r.update()
 }
 
@@ -93,7 +105,7 @@ impl CrateInfo {
         let dependency = Dependency::parse_no_deprecated(
             crate_name,
             version.as_ref().map(String::as_str),
-            &source_id,
+            source_id,
         )?;
 
         let registry_name = format!(
@@ -103,9 +115,11 @@ impl CrateInfo {
         );
 
         let (package, manifest, crate_file) = {
+            let lock = config.acquire_package_cache_lock()?;
             let mut registry = PackageRegistry::new(&config)?;
             registry.lock_patches();
             let summaries = fetch_candidates(&mut registry, &dependency)?;
+            drop(lock);
             let pkgids = summaries
                 .into_iter()
                 .map(|s| s.package_id().clone())
@@ -121,14 +135,13 @@ impl CrateInfo {
                 )
             })?;
             let pkgset = registry.get(pkgids.as_slice())?;
-            let package = pkgset.get_one(pkgid)?;
+            let package = pkgset.get_one(*pkgid)?;
             let manifest = package.manifest();
             let filename = format!("{}-{}.crate", pkgid.name(), pkgid.version());
-            let crate_file = config.registry_cache_path().join(&registry_name).open_ro(
-                &filename,
-                &config,
-                &filename,
-            )?;
+            let crate_file = config
+                .registry_cache_path()
+                .join(&registry_name)
+                .open_ro(&filename, &config, &filename)?;
             (package.clone(), manifest.clone(), crate_file)
         };
 
@@ -138,6 +151,8 @@ impl CrateInfo {
             crate_file: crate_file,
             config: config,
             source_id: source_id,
+            excludes: vec![],
+            includes: vec![],
         })
     }
 
@@ -154,7 +169,7 @@ impl CrateInfo {
     }
 
     pub fn replace_manifest(&mut self, path: &PathBuf) -> Result<&Self> {
-        if let (EitherManifest::Real(v), _) = read_manifest(path, &self.source_id, &self.config)? {
+        if let (EitherManifest::Real(v), _) = read_manifest(path, self.source_id, &self.config)? {
             self.manifest = v;
         }
         Ok(self)
@@ -164,7 +179,7 @@ impl CrateInfo {
         self.manifest.summary().checksum()
     }
 
-    pub fn package_id(&self) -> &PackageId {
+    pub fn package_id(&self) -> PackageId {
         self.manifest.summary().package_id()
     }
 
@@ -188,19 +203,35 @@ impl CrateInfo {
         self.manifest.dependencies()
     }
 
-    pub fn all_dependencies_and_features(&self) ->
-        BTreeMap<&str,                  // name of feature / optional dependency,
-                                        // or "" for the base package w/ no default features, guaranteed to be in the map
-                (Vec<&str>,             // dependencies: other features (of the current package)
-                 Vec<Dependency>)>      // dependencies: other packages
+    pub fn dev_dependencies(&self) -> Vec<Dependency> {
+        use cargo::core::dependency::Kind;
+        let mut deps = vec![];
+        for dep in self.dependencies() {
+            if dep.kind() == Kind::Development {
+                deps.push(dep.clone())
+            }
+        }
+        deps
+    }
+
+    pub fn all_dependencies_and_features(
+        &self,
+    ) -> BTreeMap<
+        &str, // name of feature / optional dependency,
+        // or "" for the base package w/ no default features, guaranteed to be in the map
+        (
+            Vec<&str>, // dependencies: other features (of the current package)
+            Vec<Dependency>,
+        ),
+    > // dependencies: other packages
     {
         use cargo::core::dependency::Kind;
 
-        let mut deps_by_name : BTreeMap<&str, Vec<&Dependency>> = BTreeMap::new();
+        let mut deps_by_name: BTreeMap<&str, Vec<&Dependency>> = BTreeMap::new();
         for dep in self.dependencies() {
             // we treat build-dependencies also as dependencies in Debian
             if dep.kind() != Kind::Development {
-                let s = dep.package_name().as_str();
+                let s = dep.name_in_toml().as_str();
                 deps_by_name.entry(s).or_default().push(dep);
             }
         }
@@ -212,20 +243,19 @@ impl CrateInfo {
         for (feature, deps) in self.manifest.summary().features() {
             let mut feature_deps = vec![""];
             // always need "", because in dh-cargo we symlink /usr/share/doc/{$feature => $main} pkg
-            let mut other_deps : Vec<Dependency> = Vec::new();
+            let mut other_deps: Vec<Dependency> = Vec::new();
             for dep in deps {
                 use self::FeatureValue::*;
                 match dep {
                     // another feature is a dependency
-                    Feature(dep_feature) =>
-                        feature_deps.push(dep_feature),
+                    Feature(dep_feature) => feature_deps.push(dep_feature),
                     // another package is a dependency
                     Crate(dep_name) => {
                         // unwrap is ok, valid Cargo.toml files must have this
                         for &dep in deps_by_name.get(dep_name.as_str()).unwrap() {
                             other_deps.push(dep.clone());
                         }
-                    },
+                    }
                     // another package is a dependency
                     CrateFeature(dep_name, dep_feature) => {
                         // unwrap is ok, valid Cargo.toml files must have this
@@ -242,11 +272,12 @@ impl CrateInfo {
         }
 
         // calculate dependencies of this crate's "optional dependencies", since they are also features
-        let mut deps_required : Vec<Dependency> = Vec::new();
+        let mut deps_required: Vec<Dependency> = Vec::new();
         for deps in deps_by_name.values() {
             for &dep in deps {
                 if dep.is_optional() {
-                    features_with_deps.insert(&dep.package_name().as_str(), (vec![""], vec![dep.clone()]));
+                    features_with_deps
+                        .insert(&dep.name_in_toml().as_str(), (vec![""], vec![dep.clone()]));
                 } else {
                     deps_required.push(dep.clone())
                 }
@@ -264,10 +295,11 @@ impl CrateInfo {
         features_with_deps
     }
 
-    pub fn feature_all_deps<'a>(&self,
-            features_with_deps: &'a BTreeMap<&str, (Vec<&str>, Vec<Dependency>)>,
-            feature: &str)
-            -> (Vec<&'a str>, Vec<Dependency>) {
+    pub fn feature_all_deps<'a>(
+        &self,
+        features_with_deps: &'a BTreeMap<&str, (Vec<&str>, Vec<Dependency>)>,
+        feature: &str,
+    ) -> (Vec<&'a str>, Vec<Dependency>) {
         let mut all_features = Vec::new();
         let mut all_deps = Vec::new();
         let &(ref ff, ref dd) = features_with_deps.get(feature).unwrap();
@@ -277,34 +309,64 @@ impl CrateInfo {
             let (ff1, dd1) = self.feature_all_deps(&features_with_deps, f);
             all_features.extend(ff1);
             all_deps.extend(dd1);
-        };
+        }
         (all_features, all_deps)
     }
 
-    // Note: this mutates features_with_deps so you need to run e.g.
-    // feature_all_deps before calling this.
-    pub fn calculate_provides<'a>(&self,
-            features_with_deps: &mut BTreeMap<&'a str, (Vec<&'a str>, Vec<Dependency>)>)
-            -> BTreeMap<&'a str, Vec<&'a str>> {
+    // Calculate Provides: in an attempt to reduce the number of binaries.
+    //
+    // Note: this mutates features_with_deps so you MUST run e.g.
+    // feature_all_deps *before* calling this.
+    //
+    // The algorithm is very simple and incomplete. e.g. it does not, yet
+    // simplify things like:
+    //   f1 depends on f2, f3
+    //   f2 depends on f4
+    //   f3 depends on f4
+    // into
+    //   f4 provides f1, f2, f3
+    pub fn calculate_provides<'a>(
+        &self,
+        features_with_deps: &mut BTreeMap<&'a str, (Vec<&'a str>, Vec<Dependency>)>,
+    ) -> BTreeMap<&'a str, Vec<&'a str>> {
+        // If any features have duplicate dependencies, deduplicate them by
+        // making all of the subsequent ones depend on the first one.
+        let mut features_rev_deps = BTreeMap::new();
+        for (&f, dep) in features_with_deps.iter() {
+            if !features_rev_deps.contains_key(dep) {
+                features_rev_deps.insert(dep.clone(), vec![]);
+            }
+            features_rev_deps.get_mut(dep).unwrap().push(f);
+        }
+        for (_, ff) in features_rev_deps.into_iter() {
+            let f0 = ff[0];
+            for f in &ff[1..] {
+                features_with_deps.insert(f, (vec!["", f0], vec![]));
+            }
+        }
+
+        // Calculate provides by following 0- or 1-length dependency lists.
         let mut provides = BTreeMap::new();
         let mut provided = Vec::new();
-        // the below is very simple and incomplete. e.g. it does not,
-        // but could be improved to, simplify things like:
-        // f1 depends on f2, f3
-        // f2 depends on f4
-        // f3 depends on f4
-        for (&f, &(ref ff, ref dd)) in features_with_deps.iter() {
+        for (&f, (ref ff, ref dd)) in features_with_deps.iter() {
+            //debcargo_info!("provides considering: {:?}", &f);
             if !dd.is_empty() {
                 continue;
             }
             assert!(!ff.is_empty() || f == "");
             let k = if ff.len() == 1 {
-                *ff.get(0).unwrap() // it's just ""
+                // if A depends only on no-default-features (""), then
+                // no-default-features provides A.
+                assert!(ff[0] == "");
+                ff[0]
             } else if ff.len() == 2 {
-                *ff.get(1).unwrap()
+                // if A depends on a single feature B, then B provides A.
+                assert!(ff[0] == "");
+                ff[1]
             } else {
                 continue;
             };
+            //debcargo_info!("provides still considering: {:?}", &f);
             if !provides.contains_key(k) {
                 provides.insert(k, vec![]);
             }
@@ -312,15 +374,20 @@ impl CrateInfo {
             provided.push(f);
         }
 
+        //debcargo_info!("provides-internal: {:?}", &provides);
+        //debcargo_info!("provided-internal: {:?}", &provided);
         for p in provided {
             features_with_deps.remove(p);
         }
 
-        features_with_deps.keys().map(|k| {
-            let mut pp = traverse_depth(&provides, k);
-            pp.sort();
-            (*k, pp)
-        }).collect::<BTreeMap<_, _>>()
+        features_with_deps
+            .keys()
+            .map(|k| {
+                let mut pp = traverse_depth(&provides, k);
+                pp.sort();
+                (*k, pp)
+            })
+            .collect::<BTreeMap<_, _>>()
     }
 
     pub fn is_lib(&self) -> bool {
@@ -369,10 +436,10 @@ impl CrateInfo {
         // See `man uscan` description of @ANY_VERSION@ on how these
         // regex patterns were built.
         match *self.package_id().version() {
-            Version { major: 0, minor, .. } =>
-                format!("[-_]?(0\\.{}\\.\\d[\\-+\\.:\\~\\da-zA-Z]*)", minor),
-            Version { major, .. } =>
-                format!("[-_]?({}\\.\\d[\\-+\\.:\\~\\da-zA-Z]*)", major),
+            Version {
+                major: 0, minor, ..
+            } => format!("[-_]?(0\\.{}\\.\\d[\\-+\\.:\\~\\da-zA-Z]*)", minor),
+            Version { major, .. } => format!("[-_]?({}\\.\\d[\\-+\\.:\\~\\da-zA-Z]*)", major),
         }
     }
 
@@ -384,14 +451,20 @@ impl CrateInfo {
                 .replace("\n\n", "\r")
                 .replace("\n", " ")
                 .replace("\r", "\n")
-                .trim().to_string();
+                .trim()
+                .to_string();
             // Trim off common prefixes
-            let re = Regex::new(&format!(r"^(?i)({}|This(\s+\w+)?)(\s*,|\s+is|\s+provides)\s+",
-                self.package_id().name())).unwrap();
+            let re = Regex::new(&format!(
+                r"^(?i)({}|This(\s+\w+)?)(\s*,|\s+is|\s+provides)\s+",
+                self.package_id().name()
+            ))
+            .unwrap();
             description = re.replace(&description, "").to_string();
             let re = Regex::new(r"^(?i)(a|an|the)\s+").unwrap();
             description = re.replace(&description, "").to_string();
-            let re = Regex::new(r"^(?i)(rust\s+)?(implementation|library|tool|crate)\s+(of|to|for)\s+").unwrap();
+            let re =
+                Regex::new(r"^(?i)(rust\s+)?(implementation|library|tool|crate)\s+(of|to|for)\s+")
+                    .unwrap();
             description = re.replace(&description, "").to_string();
 
             // https://stackoverflow.com/questions/38406793/why-is-capitalizing-the-first-letter-of-a-string-so-convoluted-in-rust
@@ -408,7 +481,7 @@ impl CrateInfo {
             let p2 = description.find(". ");
             match p1.into_iter().chain(p2.into_iter()).min() {
                 Some(p) => {
-                    let s = description[..p].trim_right_matches('.').to_string();
+                    let s = description[..p].trim_end_matches('.').to_string();
                     let d = description[p + 1..].trim();
                     if d.is_empty() {
                         (Some(s), None)
@@ -416,7 +489,7 @@ impl CrateInfo {
                         (Some(s), Some(d.to_string()))
                     }
                 }
-                None => (Some(description.trim_right_matches('.').to_string()), None),
+                None => (Some(description.trim_end_matches('.').to_string()), None),
             }
         } else {
             (None, None)
@@ -425,36 +498,67 @@ impl CrateInfo {
         (summary, description)
     }
 
-    pub fn filter_path(excludes: &Vec<Pattern>, path: &Path) -> bool {
-        // Filter out static libraries, to avoid needing to patch all the winapi crates to remove
-        // import libraries.
-        if excludes.iter().any(|p| p.matches_path(path)) {
-            return true
+    pub fn set_includes_excludes(
+        &mut self,
+        excludes: Option<&Vec<String>>,
+        includes: Option<&Vec<String>>,
+    ) {
+        self.excludes = vec_opt_iter(excludes)
+            .map(|x| Pattern::new(&("*/".to_owned() + x)).unwrap())
+            .collect::<Vec<_>>();
+        self.includes = vec_opt_iter(includes)
+            .map(|x| Pattern::new(&("*/".to_owned() + x)).unwrap())
+            .collect::<Vec<_>>();
+    }
+
+    pub fn filter_path(&self, path: &Path) -> ::std::result::Result<bool, String> {
+        if self.excludes.iter().any(|p| p.matches_path(path)) {
+            return Ok(true);
         }
-        match path.extension() {
-            Some(ext) => if ext == "a" {
-                true
-            } else {
-                if ext == "c" {
-                    debcargo_warn!("Suspicious file, maybe should be excluded: {:?}", path);
+        let suspicious = match path.extension() {
+            Some(ext) => {
+                if ext == "c" || ext == "a" {
+                    true
+                } else {
+                    false
                 }
-                false
-            },
+            }
             _ => false,
+        };
+        if suspicious {
+            if self.includes.iter().any(|p| p.matches_path(path)) {
+                debcargo_info!("Suspicious file, on whitelist so ignored: {:?}", path);
+                Ok(false)
+            } else {
+                Err(format!(
+                    "Suspicious file, should probably be excluded: {:?}",
+                    path
+                ))
+            }
+        } else {
+            Ok(false)
         }
     }
 
-    pub fn extract_crate(&self, path: &Path, excludes: &Vec<Pattern>) -> Result<bool> {
+    pub fn extract_crate(&self, path: &Path) -> Result<bool> {
         let mut archive = Archive::new(GzDecoder::new(self.crate_file.file()));
-        let tempdir = tempfile::Builder::new().prefix("debcargo").tempdir_in(".")?;
+        let tempdir = tempfile::Builder::new()
+            .prefix("debcargo")
+            .tempdir_in(".")?;
         let mut source_modified = false;
         let mut last_mtime = 0;
+        let mut err = vec![];
 
         for entry in archive.entries()? {
             let mut entry = entry?;
-            if Self::filter_path(excludes, &(entry.path()?)) {
-                source_modified = true;
-                continue;
+            match self.filter_path(&(entry.path()?)) {
+                Err(e) => err.push(e),
+                Ok(r) => {
+                    if r {
+                        source_modified = true;
+                        continue;
+                    }
+                }
             }
 
             if !entry.unpack_in(tempdir.path())? {
@@ -467,6 +571,14 @@ impl CrateInfo {
                 }
             }
         }
+        if !err.is_empty() {
+            for e in err {
+                debcargo_warn!("{}", e);
+            }
+            debcargo_bail!(
+                "Suspicious files detected, aborting. Ask on #debian-rust if you are stuck."
+            )
+        }
 
         let entries = tempdir.path().read_dir()?.collect::<io::Result<Vec<_>>>()?;
         if entries.len() != 1 || !entries[0].file_type()?.is_dir() {
@@ -490,19 +602,28 @@ impl CrateInfo {
 
         // Ensure that Cargo.toml is in standard form, e.g. does not contain
         // path dependencies, so can be built standalone (see #4030).
-        let registry_toml = self.package().to_registry_toml(&Config::default()?)?;
+        let registry_toml = self.package().to_registry_toml(&self.config)?;
         let mut actual_toml = String::new();
         let toml_path = path.join("Cargo.toml");
         fs::File::open(&toml_path)?.read_to_string(&mut actual_toml)?;
 
-        if actual_toml != registry_toml {
+        if !actual_toml.contains("AUTOMATICALLY GENERATED BY CARGO") {
+            // This logic should only fire for old crates, and that's what the
+            // if-conditional is supposed to check; modern versions of cargo
+            // already do this before uploading the crate and we shouldn't need
+            // to handle it specially.
             let old_toml_path = path.join("Cargo.toml.orig");
-            fs::rename(&toml_path, &old_toml_path)?;
+            fs::copy(&toml_path, &old_toml_path)?;
             fs::OpenOptions::new()
-                .create(true)
                 .write(true)
+                .truncate(true)
                 .open(&toml_path)?
                 .write_all(registry_toml.as_bytes())?;
+            debcargo_info!(
+                "Rewrote {:?} to canonical form\nOld backed up as {:?}",
+                &toml_path,
+                &old_toml_path,
+            );
             source_modified = true;
             // avoid lintian errors about package-contains-ancient-file
             // TODO: do we want to do this for unmodified tarballs? it would
